"""Regular expressions for Emoji Definitions

ref: http://www.unicode.org/reports/tr51/#Definitions
"""

import re
from enum import Enum
from typing import Pattern, Mapping, MutableMapping

from .character import (
    EMOJI_KEYCAP,
    EMOJI_PRESENTATION_SELECTOR,
    REGIONAL_INDICATORS,
    TAGS,
    TEXT_PRESENTATION_SELECTOR,
    ZWJ,
    EmojiCharacter,
    EmojiCharProperty,
)
from .utils import code_point_to_regex

__all__ = [
    "EMOJI_PATTERNS",
    "QualifiedType",
    "detect_qualified",
    "is_default_emoji_presentation_character",
    "is_default_text_presentation_character",
    "is_emoji_character",
    "is_emoji_core_sequence",
    "is_emoji_flag_sequence",
    "is_emoji_keycap_sequence",
    "is_emoji_modifier",
    "is_emoji_modifier_base",
    "is_emoji_modifier_sequence",
    "is_emoji_presentation_selector",
    "is_emoji_presentation_sequence",
    "is_emoji_sequence",
    "is_emoji_tag_sequence",
    "is_emoji_zwj_element",
    "is_emoji_zwj_sequence",
    "is_regional_indicator",
    "is_tag_base",
    "is_tag_spec",
    "is_tag_term",
    "is_text_presentation_selector",
    "is_text_presentation_sequence",
    "is_qualified_emoji_character",
]


class QualifiedType(Enum):
    FULLY_QUALIFIED = "fully-qualified"
    MINIMALLY_QUALIFIED = "minimally-qualified"
    UNQUALIFIED = "unqualified"


def make_regex_dict() -> Mapping[str, str]:
    d: MutableMapping[str, str] = {}

    d["EMOJI_CHARACTER"] = (
        r"[" + "".join(m.regex for m in EmojiCharacter.values() if EmojiCharProperty.EMOJI in m.properties) + r"]"
    )

    d["DEFAULT_EMOJI_PRESENTATION_CHARACTER"] = (
        r"[" + "".join(m.regex for m in EmojiCharacter.values() if EmojiCharProperty.EPRES in m.properties) + r"]"
    )

    d["DEFAULT_TEXT_PRESENTATION_CHARACTER"] = (
        r"[" + "".join(m.regex for m in EmojiCharacter.values() if EmojiCharProperty.EPRES not in m.properties) + r"]"
    )
    d["TEXT_PRESENTATION_SELECTOR"] = code_point_to_regex(TEXT_PRESENTATION_SELECTOR)
    d["TEXT_PRESENTATION_SEQUENCE"] = r"({EMOJI_CHARACTER}{TEXT_PRESENTATION_SELECTOR})".format(**d)
    d["EMOJI_PRESENTATION_SELECTOR"] = code_point_to_regex(EMOJI_PRESENTATION_SELECTOR)
    d["EMOJI_PRESENTATION_SEQUENCE"] = r"({EMOJI_CHARACTER}{EMOJI_PRESENTATION_SELECTOR})".format(**d)
    d["EMOJI_MODIFIER"] = (
        r"[" + "".join(m.regex for m in EmojiCharacter.values() if EmojiCharProperty.EMOD in m.properties) + r"]"
    )
    d["EMOJI_MODIFIER_BASE"] = (
        r"[" + "".join(m.regex for m in EmojiCharacter.values() if EmojiCharProperty.EBASE in m.properties) + r"]"
    )
    d["EMOJI_MODIFIER_SEQUENCE"] = r"({EMOJI_MODIFIER_BASE}{EMOJI_MODIFIER})".format(**d)
    d["REGIONAL_INDICATOR"] = (
        r"[" + code_point_to_regex(REGIONAL_INDICATORS[0]) + r"-" + code_point_to_regex(REGIONAL_INDICATORS[-1]) + r"]"
    )
    d["EMOJI_FLAG_SEQUENCE"] = r"({REGIONAL_INDICATOR}{REGIONAL_INDICATOR})".format(**d)
    d["TAG_BASE"] = r"({EMOJI_CHARACTER}|{EMOJI_MODIFIER_SEQUENCE}|{EMOJI_PRESENTATION_SEQUENCE})".format(**d)
    d["TAG_SPEC"] = r"[" + code_point_to_regex(TAGS[0]) + r"-" + code_point_to_regex(TAGS[-2]) + r"]"
    d["TAG_TERM"] = code_point_to_regex(TAGS[-1])
    d["EMOJI_TAG_SEQUENCE"] = r"({TAG_BASE}{TAG_SPEC}{TAG_TERM})".format(**d)
    d["EMOJI_KEYCAP_SEQUENCE"] = r"([0-9#*]{}{})".format(
        *(code_point_to_regex(x) for x in (EMOJI_PRESENTATION_SELECTOR, EMOJI_KEYCAP))
    )
    d["EMOJI_CORE_SEQUENCE"] = (
        r"("
        r"{EMOJI_CHARACTER}"
        r"|{EMOJI_PRESENTATION_SEQUENCE}"
        r"|{EMOJI_KEYCAP_SEQUENCE}"
        r"|{EMOJI_MODIFIER_SEQUENCE}"
        r"|{EMOJI_FLAG_SEQUENCE}"
        r")".format(**d)
    )
    d["EMOJI_ZWJ_ELEMENT"] = r"({EMOJI_CHARACTER}|{EMOJI_PRESENTATION_SEQUENCE}|{EMOJI_MODIFIER_SEQUENCE})".format(**d)
    d["EMOJI_ZWJ_SEQUENCE"] = r"({EMOJI_ZWJ_ELEMENT}({0}{EMOJI_ZWJ_ELEMENT})+)".format(code_point_to_regex(ZWJ), **d)
    d["EMOJI_SEQUENCE"] = r"({EMOJI_CORE_SEQUENCE}|{EMOJI_ZWJ_SEQUENCE}|{EMOJI_TAG_SEQUENCE})".format(**d)

    return d


EMOJI_PATTERNS: Mapping[str, Pattern[str]] = {k: re.compile(v) for k, v in make_regex_dict().items()}


def is_emoji_character(c: str) -> bool:
    """detect emoji character

    A character that has the Emoji property.

    ::

        emoji_character := \\p{Emoji}

    - These characters are recommended for use as emoji.

    See also:
        https://unicode.org/reports/tr51/#Emoji_Characters
    """
    _c = chr(ord(c))
    return EMOJI_PATTERNS["EMOJI_CHARACTER"].fullmatch(_c) is not None


def is_default_emoji_presentation_character(c: str) -> bool:
    """detect default emoji presentation character

    A character that, by default, should appear with an emoji presentation, rather than a text presentation.

    ::

        default_emoji_presentation_character := \\p{Emoji_Presentation}

    - These characters have the Emoji_Presentation property.

    See also:
        https://unicode.org/reports/tr51/#def_emoji_presentation
    """
    _c = chr(ord(c))
    return EMOJI_PATTERNS["DEFAULT_EMOJI_PRESENTATION_CHARACTER"].fullmatch(_c) is not None


def is_default_text_presentation_character(c: str) -> bool:
    """detect default text presentation character

    A character that, by default, should appear with a text presentation, rather than an emoji presentation.

    See also:
        https://unicode.org/reports/tr51/#def_text_presentation_sequence
    """
    _c = chr(ord(c))
    return EMOJI_PATTERNS["DEFAULT_TEXT_PRESENTATION_CHARACTER"].fullmatch(_c) is not None


def is_text_presentation_selector(c: str) -> bool:
    """detect text presentation selector

    The character U+FE0E VARIATION SELECTOR-15 (VS15), used to request a text presentation for an emoji character. (Also known as text variation selector in prior versions of this specification.)

    See also:
        https://unicode.org/reports/tr51/#def_emoji_presentation_selector
    """
    return EMOJI_PATTERNS["TEXT_PRESENTATION_SELECTOR"].fullmatch(c) is not None


def is_text_presentation_sequence(s: str) -> bool:
    """detect text presentation selector

    The character U+FE0E VARIATION SELECTOR-15 (VS15), used to request a text presentation for an emoji character. (Also known as text variation selector in prior versions of this specification.)

    See also:
        https://unicode.org/reports/tr51/#def_text_presentation_sequence
    """
    return EMOJI_PATTERNS["TEXT_PRESENTATION_SEQUENCE"].fullmatch(s) is not None


def is_emoji_presentation_selector(c: str) -> bool:
    """detect emoji presentation selector

    The character U+FE0F VARIATION SELECTOR-16 (VS16), used to request an emoji presentation for an emoji character. (Also known as emoji variation selector in prior versions of this specification.)

    See also:
        https://unicode.org/reports/tr51/#def_emoji_presentation_selector
    """
    return EMOJI_PATTERNS["EMOJI_PRESENTATION_SELECTOR"].fullmatch(c) is not None


def is_emoji_presentation_sequence(s: str) -> bool:
    """detect emoji presentation sequence

    A variation sequence consisting of an emoji character followed by a emoji presentation selector.

    ::

        emoji_presentation_sequence := emoji_character emoji_presentation_selector

    - The only valid emoji presentation sequences are those listed in emoji-variation-sequences.txt

    See also:
        https://unicode.org/reports/tr51/#def_emoji_presentation_sequence
    """
    return EMOJI_PATTERNS["EMOJI_PRESENTATION_SEQUENCE"].fullmatch(s) is not None


def is_emoji_modifier(c: str) -> bool:
    """detect emoji modifier

    A character that can be used to modify the appearance of a preceding emoji in an emoji modifier sequence.

    See also:
        https://unicode.org/reports/tr51/#def_emoji_modifier
    """
    _c = chr(ord(c))
    return EMOJI_PATTERNS["EMOJI_MODIFIER"].fullmatch(_c) is not None


def is_emoji_modifier_base(c: str) -> bool:
    """Detect emoji modifier base

    A character whose appearance can be modified by a subsequent emoji modifier in an emoji modifier sequence.

    See also:
        https://unicode.org/reports/tr51/#def_emoji_modifier_base
    """
    _c = chr(ord(c))
    return EMOJI_PATTERNS["EMOJI_MODIFIER_BASE"].fullmatch(_c) is not None


def is_emoji_modifier_sequence(s: str) -> bool:
    """Detect emoji modifier sequence

    A sequence of the following form::

        emoji_modifier_sequence := emoji_modifier_base emoji_modifier
    """
    return EMOJI_PATTERNS["EMOJI_MODIFIER_SEQUENCE"].fullmatch(s) is not None


def is_regional_indicator(s: str) -> bool:
    """A singleton Regional Indicator character is not a well-formed emoji flag sequence."""
    return EMOJI_PATTERNS["REGIONAL_INDICATOR"].fullmatch(s) is not None


def is_emoji_flag_sequence(s: str) -> bool:
    """emoji flag sequence — A sequence of two Regional Indicator characters,
    where the corresponding ASCII characters are valid region sequences as specified by `Unicode region subtags <https://www.unicode.org/reports/tr35/#unicode_region_subtag>`_ in [`CLDR <https://www.unicode.org/reports/tr51/#CLDR>`_],
    with idStatus = “regular”, “deprecated”, or “macroregion”.

    See also:
        `Annex B: Valid Emoji Flag Sequences <https://www.unicode.org/reports/tr51/#Flags>`_.

    ::

        emoji_flag_sequence :=
        regional_indicator regional_indicator

        regional_indicator := \\p{Regional_Indicator}

    A singleton Regional Indicator character is not a well-formed **emoji flag sequence**.

    See also:
        https://www.unicode.org/reports/tr51/#def_emoji_flag_sequence

    """
    return EMOJI_PATTERNS["EMOJI_FLAG_SEQUENCE"].fullmatch(s) is not None


def is_tag_base(s: str) -> bool:
    return EMOJI_PATTERNS["TAG_BASE"].fullmatch(s) is not None


def is_tag_spec(s: str) -> bool:
    return EMOJI_PATTERNS["TAG_SPEC"].fullmatch(s) is not None


def is_tag_term(c: str) -> bool:
    return EMOJI_PATTERNS["TAG_TERM"].fullmatch(c) is not None


def is_emoji_tag_sequence(s: str) -> bool:
    """emoji tag sequence (ETS) — A sequence of the following form::

        emoji_tag_sequence := tag_base tag_spec tag_end
        tag_base           := emoji_character
                            | emoji_modifier_sequence
                            | emoji_presentation_sequence
        tag_spec           := [\\x{E0020}-\\x{E007E}]+
        tag_end            := \\x{E007F}

    * The `tag_spec` consists of all characters from U+E0020 TAG SPACE to U+E007E TAG TILDE. Each tag_spec defines a particular visual variant to be applied to the tag_base character(s). Though tag_spec includes the values U+E0041 TAG LATIN CAPITAL LETTER A .. U+E005A TAG LATIN CAPITAL LETTER Z, they are not used currently and are reserved for future extensions.
    * The `tag_end` consists of the character U+E007F CANCEL TAG, and must be used to terminate the sequence.
    * A sequence of tag characters that is not part of an `emoji_tag_sequence` is not a well-formed **emoji tag sequence**.

    See also:
        https://www.unicode.org/reports/tr51/#def_emoji_tag_sequence
    """
    return EMOJI_PATTERNS["EMOJI_TAG_SEQUENCE"].fullmatch(s) is not None


def is_emoji_keycap_sequence(s: str) -> bool:
    """A sequence of the following form::

        emoji_keycap_sequence := [0-9#*] \\x{FE0F 20E3}

    * These sequences are in the emoji-sequences.txt file listed under the type_field **Emoji_Keycap_Sequence**

    See also:
        https://www.unicode.org/reports/tr51/#def_emoji_keycap_sequence
    """
    return EMOJI_PATTERNS["EMOJI_KEYCAP_SEQUENCE"].fullmatch(s) is not None


def is_emoji_core_sequence(s: str) -> bool:
    """emoji core sequence — A sequence of the following form::

        emoji_core_sequence :=
            emoji_character
        | emoji_presentation_sequence
        | emoji_keycap_sequence
        | emoji_modifier_sequence
        | emoji_flag_sequence

    See also:
        https://www.unicode.org/reports/tr51/#def_emoji_core_sequence
    """
    return EMOJI_PATTERNS["EMOJI_CORE_SEQUENCE"].fullmatch(s) is not None


def is_emoji_zwj_element(s: str) -> bool:
    """emoji ZWJ element — An element that can be used in an emoji ZWJ sequence, as follows::

        emoji_zwj_element :=
            emoji_core_sequence
        | emoji_tag_sequence

    See also:
        https://www.unicode.org/reports/tr51/#def_emoji_zwj_element
    """
    return EMOJI_PATTERNS["EMOJI_ZWJ_ELEMENT"].fullmatch(s) is not None


def is_emoji_zwj_sequence(s: str) -> bool:
    """emoji sequence — A core sequence, tag sequence, or ZWJ sequence, as follows::

        emoji_sequence :=
            emoji_core_sequence
        | emoji_zwj_sequence
        | emoji_tag_sequence

    Note:
        all emoji sequences are single grapheme clusters: there is never a grapheme cluster boundary within an emoji sequence.
        This affects editing operations, such as cursor movement or deletion, as well as word break, line break, and so on.
        For more information, see [`UAX29 <https://www.unicode.org/reports/tr51/#UAX29>`_].

    See also:
        https://www.unicode.org/reports/tr51/#def_emoji_sequence
    """
    return EMOJI_PATTERNS["EMOJI_ZWJ_SEQUENCE"].fullmatch(s) is not None


def is_emoji_sequence(s: str) -> bool:
    """emoji sequence — A core sequence, tag sequence, or ZWJ sequence, as follows::

        emoji_sequence :=
            emoji_core_sequence
        | emoji_zwj_sequence
        | emoji_tag_sequence

    Note:
        all emoji sequences are single grapheme clusters: there is never a grapheme cluster boundary within an emoji sequence.
        This affects editing operations, such as cursor movement or deletion, as well as word break, line break, and so on.
        For more information, see [`UAX29 <https://www.unicode.org/reports/tr51/#UAX29>`_].

    See also:
        https://www.unicode.org/reports/tr51/#def_emoji_sequence
    """
    return EMOJI_PATTERNS["EMOJI_SEQUENCE"].fullmatch(s) is not None


def is_qualified_emoji_character(s: str, i: int) -> bool:
    """An emoji character in a string that

    - (a) has default emoji presentation or
    - (b) is the first character in an emoji modifier sequence or
    - (c) is not a default emoji presentation character, but is the first character in an emoji presentation sequence.

    Args:
        s: the string where the character in it
        i: index of the character in the string to check if qualified

    See also:
        http://www.unicode.org/reports/tr51/#def_qualified_emoji_character
    """
    c = s[i]
    if not is_emoji_character(c):
        return False
    if is_default_emoji_presentation_character(c):  # default emoji presentation
        return True
    if EMOJI_PATTERNS["EMOJI_MODIFIER_SEQUENCE"].match(s[i:]):  # first character in an emoji modifier sequence
        return True
    if EMOJI_PATTERNS["EMOJI_PRESENTATION_SEQUENCE"].match(s[i:]):  # first character in an emoji presentation sequence
        return True
    return False


def detect_qualified(s: str) -> QualifiedType:
    """Detect qualified type of emoji string

    - qualified emoji character — An emoji character in a string that

        - (a) has default emoji presentation or
        - (b) is the first character in an emoji modifier sequence or
        - (c) is not a default emoji presentation character, but is the first character in an emoji presentation sequence.

    - fully-qualified emoji — A qualified emoji character, or an emoji sequence in which each emoji character is qualified.
    - minimally-qualified emoji — An emoji sequence in which the first character is qualified but the sequence is not fully qualified.
    - unqualified emoji — An emoji that is neither fully-qualified nor minimally qualified.

    Args:
        s: string to detect
    """
    if not s:
        raise ValueError("Argument `s` should not be empty or null")
    if is_qualified_emoji_character(s, 0):
        n = len(s)
        if n == 1:
            return QualifiedType.FULLY_QUALIFIED
        if all(is_qualified_emoji_character(s, i) for i in range(1, n) if is_emoji_character(s[i])):
            return QualifiedType.FULLY_QUALIFIED
        return QualifiedType.MINIMALLY_QUALIFIED
    return QualifiedType.UNQUALIFIED



from __future__ import annotations

import re
import sys
from enum import Enum
from typing import Iterable, Iterator, MutableSequence, Optional, Sequence, Tuple, Union, final

if sys.version_info < (3, 11):  # pragma: no cover
    from typing_extensions import Self
else:  # pragma: no cover
    from typing import Self

from .types import BaseDictContainer
from .utils import code_point_to_regex, emoji_data_lines

__all__ = [
    "EmojiCharProperty",
    "EmojiCharacter",
    "TEXT_PRESENTATION_SELECTOR",
    "EMOJI_PRESENTATION_SELECTOR",
    "EMOJI_KEYCAP",
    "REGIONAL_INDICATORS",
    "TAGS",
    "ZWJ",
]


TEXT_PRESENTATION_SELECTOR = 0xFE0E
"""The character U+FE0E VARIATION SELECTOR-15 (VS15), used to request a text presentation for an emoji character.
(Also known as text variation selector in prior versions of this specification.)
"""

EMOJI_PRESENTATION_SELECTOR = 0xFE0F
"""The character U+FE0F VARIATION SELECTOR-16 (VS16), used to request an emoji presentation for an emoji character.
(Also known as emoji variation selector in prior versions of this specification.)
"""

EMOJI_KEYCAP = 0x20E3
"""A sequence of the following form::

    emoji_keycap_sequence := [0-9#*] \\x{FE0F 20E3}

- These sequences are in the `emoji-sequences.txt` file listed under the type_field ``Emoji_Keycap_Sequence``
"""

ZWJ = 0x200D
"""An emoji sequence with at least one joiner character.
"""

REGIONAL_INDICATORS = list(range(0x1F1E6, 0x1F1FF + 1))
"""regional indicators"""

TAGS = list(range(0xE0020, 0xE007F + 1))
"""tags"""


class EmojiCharProperty(Enum):
    """Emoji Character Properties

    character properties are available for emoji characters.

    See also:
        http://www.unicode.org/reports/tr51/#Emoji_Properties
    """

    EMOJI = "Emoji"
    """for characters that are emoji"""

    EPRES = "Emoji_Presentation"
    """ for characters that have emoji presentation by default"""

    EMOD = "Emoji_Modifier"
    """for characters that are emoji modifiers"""

    EBASE = "Emoji_Modifier_Base"
    """for characters that can serve as a base for emoji modifiers"""

    ECOMP = "Emoji_Component"
    """for characters used in emoji sequences that normally do not appear on emoji keyboards as separate choices, such as keycap base characters or Regional_Indicator characters.

    All characters in emoji sequences are either Emoji or Emoji_Component.
    Implementations must not, however, assume that all Emoji_Component characters are also Emoji.
    There are some non-emoji characters that are used in various emoji sequences, such as tag characters and ZWJ.
    """

    EXTPICT = "Extended_Pictographic"
    """for characters that are used to future-proof segmentation.

    The Extended_Pictographic characters contain all the Emoji characters except for some Emoji_Component characters.
    """


class MetaClass(BaseDictContainer[int, "EmojiCharacter"]):
    pass


@final
class EmojiCharacter(metaclass=MetaClass):  # pyright: ignore[reportGeneralTypeIssues]
    """emoji character — A character that has the Emoji property.

    These characters are recommended for use as emoji.

    See also:
        http://www.unicode.org/reports/tr51/#Emoji_Characters
    """

    def __init__(
        self,
        code_point: int,
        properties: Union[EmojiCharProperty, Iterable[EmojiCharProperty], None] = None,
        version: Optional[str] = None,
        description: Optional[str] = None,
    ):
        self._code_point = code_point
        self._string = chr(self._code_point)
        self._regex = code_point_to_regex(code_point)
        #
        self._properties: MutableSequence[EmojiCharProperty]
        if properties is None:
            self._properties = []
        elif isinstance(properties, EmojiCharProperty):
            self._properties = [properties]
        elif isinstance(properties, Iterable):
            self._properties = list(properties)
        else:
            raise TypeError(f"{type(properties)}")
        #
        self._version = version or ""
        self._description = description or ""

    def __str__(self):
        return self._string

    def __repr__(self):
        return "<{} code_point={} char={!r} version={!r} description={!r}>".format(
            type(self).__name__, self.code_point_string, self.string, self.version, self.description
        )

    _comment_split_regex = re.compile(r"\[\d+\]\s*\(.*\)")

    _initialed = False

    @classmethod
    def initial(cls):
        """Initial the class

        Load Emoji Characters and it's properties from package data file into class internal dictionary
        """
        if cls._initialed:
            return
        for content, comment in emoji_data_lines("emoji-data.txt"):
            cps, property_text = (part.strip() for part in content.split(";", 1))
            cps_parts = cps.split("..", 1)
            property_ = EmojiCharProperty(property_text)
            version, description = (s.strip() for s in cls._comment_split_regex.split(comment, maxsplit=1))
            for cp in range(int(cps_parts[0], 16), 1 + int(cps_parts[-1], 16)):
                try:
                    inst = cls[cp]
                except KeyError:
                    cls[cp] = cls(cp, property_, version, description)
                else:
                    inst._add_property(property_)
        for cp in (TEXT_PRESENTATION_SELECTOR, EMOJI_PRESENTATION_SELECTOR, EMOJI_KEYCAP):
            if cp not in cls:
                cls[cp] = cls(cp, [])
        # OK!
        cls._initialed = True

    @classmethod
    def release(cls):
        if not cls._initialed:
            return
        keys = list(cls)
        for k in keys:
            del cls[k]
        cls._initialed = False

    @classmethod
    def items(cls) -> Iterator[Tuple[int, Self]]:
        """Returns an iterator of all code-point -> emoji-character pairs of the class"""
        return ((k, cls[k]) for k in cls)  # pyright: ignore[reportReturnType]

    @classmethod
    def keys(cls) -> Iterator[int]:
        """Returns an iterator of each emoji-character's key code-point of the class"""
        yield from cls

    @classmethod
    def values(cls) -> Iterator[Self]:
        """Returns an iterator of all emoji-characters of the class"""
        return (cls[k] for k in cls)  # pyright: ignore[reportReturnType]

    def _add_property(self, val: EmojiCharProperty):
        if val not in self._properties:
            self._properties.append(val)

    @property
    def code_point(self) -> int:
        """Unicode integer value of the emoji-characters"""
        return self._code_point

    @property
    def code_point_string(self) -> str:
        """Unicode style hex string of the emoji-characters's code-point

        Example:
            ``"25FB"``
        """
        return f"{self._code_point:04X}"

    @property
    def properties(self) -> Sequence[EmojiCharProperty]:
        """Property description text of the emoji-characters"""
        return self._properties

    @property
    def version(self) -> str:
        """Version of the Emoji.

        Example:
            ``E0.0``, ``E0.6``, ``E11.0``
        """
        return self._version

    @property
    def description(self) -> str:
        """Description comment of the Emoji"""
        return self._description

    @property
    def regex(self) -> str:
        """Regular express for the emoji-characters"""
        return self._regex

    @property
    def hex(self) -> str:
        """Python style hex string of the emoji-characters's code-pint

        Example:
            ``"0x25fb"``
        """
        return hex(self._code_point)

    @property
    def string(self) -> str:
        """Emoji character string"""
        return self._string

    @classmethod
    def from_character(cls, c: str) -> Self:
        """Get :class:`EmojiCharacter` instance from a single Emoji Unicode character

        Args:
            c: Emoji character

                Note:
                    ``c`` should be a **single** unicode character, that is: ``len(c) == 1``.

        Returns:
            Instance returned from the class's internal dictionary

        Raises:
            KeyError: When character not found in the class' internal dictionary
        """
        return cls[ord(c)]  # pyright: ignore[reportReturnType]

    @classmethod
    def from_hex(cls, value: Union[int, str]) -> Self:
        """Get an :class:`EmojiCharacter` instance by Emoji Unicode integer value or it's hex string

        Args:
            value: Emoji Unicode, either integer value or hex string

        Returns:
            Instance returned from the class's internal dictionary

        Raises:
            KeyError: When code not found in the class' internal dictionary
        """
        if isinstance(value, str):
            return cls[int(value, 16)]  # pyright: ignore[reportReturnType]
        return cls[int(value)]  # pyright: ignore[reportReturnType]


EmojiCharacter.initial()



from __future__ import annotations

import sys
from typing import Union, Generator, Iterable, Tuple

if sys.version_info < (3, 9):  # pragma: no cover
    import importlib_resources
else:  # pragma: no cover
    import importlib.resources as importlib_resources


__all__ = ["code_points_to_string", "code_point_to_regex", "emoji_data_lines"]


def emoji_data_lines(data_file: str) -> Generator[Tuple[str, str], None, None]:
    ft = importlib_resources.files(__package__).joinpath("data").joinpath(data_file)
    for line in ft.read_text("utf-8").splitlines():
        line = line.strip()
        if not line:
            continue
        if line[0] in "#;":
            continue
        parts = [s.strip() for s in line.split("#", 1)]
        content = parts[0]
        try:
            comment = parts[1]
        except IndexError:
            comment = ""
        yield content.strip(), comment.strip()


def code_points_to_string(code_points: Union[int, str, Iterable[Union[int, str]]]) -> str:
    if isinstance(code_points, str):
        return "".join(chr(int(s, 16)) for s in code_points.split())
    if isinstance(code_points, Iterable):
        return "".join(chr(int(n, 16)) if isinstance(n, str) else chr(n) for n in code_points)
    return chr(code_points)


def code_point_to_regex(code_point: int) -> str:
    return rf"\U{code_point:08X}" if code_point > 0xFFFF else rf"\u{code_point:04X}"



"""
A library represents emoji characters and sequences defined in `Unicode® Technical Standard #51 Data Files <http://www.unicode.org/reports/tr51/#Data_Files_Table>`_
"""

from .character import *
from .definitions import *
from .sequence import *
from .utils import *
from ._version import __version__, __version_tuple__

EMOJI_VERSION = "15.1"
"""Starting with Version 11.0 of Unicode® Technical Standard #51(UNICODE EMOJI) specification, the repertoire of emoji characters is synchronized with the Unicode Standard, and has the same version numbering system.
"""

EMOJI_REVISION = "25"
"""Revision 25
"""



from typing import Any, Dict, Generator, Generic, MutableMapping, Tuple, Type, TypeVar

__all__ = ["BaseDictContainer"]

TK = TypeVar("TK")
TV = TypeVar("TV")


class BaseDictContainer(type, Generic[TK, TV]):
    __data__: MutableMapping[TK, TV]

    def __new__(cls, name: str, bases: Tuple[Type, ...], attrs: Dict[str, Any]):
        cls.__data__ = {}
        return super().__new__(cls, name, bases, attrs)

    def __setitem__(self, key: TK, value: TV):
        self.__data__[key] = value

    def __delitem__(self, key: TK):
        del self.__data__[key]

    def __getitem__(self, key: TK) -> TV:
        return self.__data__[key]

    def __contains__(self, key: TK) -> bool:
        return key in self.__data__

    def __iter__(self) -> Generator[TK, None, None]:
        yield from self.__data__

    def __len__(self) -> int:
        return len(self.__data__)



from __future__ import annotations

import re
import sys
from typing import ClassVar, Generator, Iterable, Iterator, Optional, Pattern, Sequence, Tuple, Union, final

if sys.version_info < (3, 11):  # pragma: no cover
    from typing_extensions import Self
else:  # pragma: no cover
    from typing import Self

from .character import EmojiCharacter
from .types import BaseDictContainer
from .utils import emoji_data_lines

__all__ = ["EmojiSequence"]

# http://www.unicode.org/reports/tr51/#Data_Files_Table
# keep the order!
DATA_FILES = ("emoji-variation-sequences.txt", "emoji-zwj-sequences.txt", "emoji-sequences.txt")


class MetaClass(BaseDictContainer[str, "EmojiSequence"]):
    pass


@final
class EmojiSequence(metaclass=MetaClass):  # pyright: ignore[reportGeneralTypeIssues]
    """Emoji and Text Presentation Sequences used to represent emoji

    See also:
        http://www.unicode.org/reports/tr51/#Emoji_Sequences
    """

    def __init__(
        self,
        code_points: Union[int, Iterable[int]],
        type_field: Optional[str] = None,
        version: Optional[str] = None,
        variation: Optional[str] = None,
        description: Optional[str] = None,
    ):
        if isinstance(code_points, Iterable):
            self._code_points = list(code_points)
        else:
            self._code_points = [code_points]
        self._string = "".join(chr(n) for n in self._code_points)
        self._characters = [EmojiCharacter.from_hex(n) for n in self._code_points]
        self._type_field = type_field or ""
        self._version = version or ""
        self._variation = variation or ""
        self._description = description or ""
        # regex
        self._regex = r""
        if not self._regex:
            self._regex = "".join(m.regex for m in self._characters)
        self._regex_pat = re.compile(self._regex)

    def __len__(self):
        return len(self._code_points)

    def __str__(self):
        return self._string

    def __repr__(self):
        return "<{} code_points={!r} string={!r} version={!r} description={!r}>".format(
            type(self).__name__, self.code_points_string, self.string, self.version, self.description
        )

    _initialed = False

    pattern: ClassVar[Pattern[str]]
    """Compiled regular expression pattern object for all-together Emoji sequences.
    """

    @classmethod
    def initial(cls):
        """Initial the class

        Load Emoji Sequences from package data file into class internal dictionary
        """
        if cls._initialed:
            return
        EmojiCharacter.initial()

        def _decode_code_points(_cps, **_kwargs):
            try:
                _cp_head, _cp_tail = _cps.split("..", 1)  # begin..end form
            except ValueError:
                _arr_cp = [int(x, 16) for x in _cps.split()]
                _seq = cls(_arr_cp, **_kwargs)
                cls[_seq.string] = _seq
            else:
                # begin..end form: A range of single char emoji-seq
                for _cp in range(int(_cp_head, 16), 1 + int(_cp_tail, 16)):
                    _seq = cls(_cp, **_kwargs)
                    cls[_seq.string] = _seq

        for fname in DATA_FILES:
            for content, comment in emoji_data_lines(fname):
                if fname in ("emoji-sequences.txt", "emoji-zwj-sequences.txt"):
                    cps, type_field, description = (part.strip() for part in content.split(";", 2))
                    version = comment.split(maxsplit=1)[0]
                    _decode_code_points(cps, type_field=type_field, version=version, description=description)
                elif fname == "emoji-variation-sequences.txt":
                    cps, variation, _ = (part.strip() for part in content.split(";", 2))
                    version, description = (x.strip() for x in comment.split(maxsplit=1))
                    version = "E" + version.lstrip("(").rstrip(")").strip()
                    description = f"{description.lower()} ({variation})"
                    _decode_code_points(cps, version=version, variation=variation, description=description)
                else:
                    raise RuntimeError(f"Invalid data file name {fname}")
        # build regex
        cls.pattern = re.compile(
            r"|".join(
                m.regex
                for m in sorted(
                    (m for m in cls.values()),
                    key=lambda x: len(x.code_points),
                    reverse=True,
                )
            )
        )
        # initialed OK
        cls._initialed = True

    @classmethod
    def release(cls):
        if not cls._initialed:
            return
        keys = list(cls)
        for k in keys:
            del cls[k]
        cls._initialed = False

    @classmethod
    def items(cls) -> Iterator[Tuple[str, Self]]:
        """Returns an iterator of all string -> emoji-sequence pairs of the class"""
        return ((k, cls[k]) for k in cls)  # pyright: ignore[reportReturnType]

    @classmethod
    def keys(cls) -> Iterator[str]:
        """Returns an iterator of each emoji-sequence's key string of the class"""
        yield from cls

    @classmethod
    def values(cls) -> Iterator[Self]:
        """Returns an iterator of all emoji-sequences of the class"""
        return (cls[k] for k in cls)  # pyright: ignore[reportReturnType]

    @classmethod
    def from_string(cls, s: str) -> Self:
        """Get an :class:`EmojiSequence` instance from string

        Args:
            s: Emoji string

        Returns:
            Instance from internal dictionary

        Raises:
            KeyError: When passed-in ``s`` not found in internal dictionary
        """
        return cls[s]  # pyright: ignore[reportReturnType]

    @classmethod
    def from_characters(cls, value: Union[EmojiCharacter, Iterable[EmojiCharacter]]) -> Self:
        """Get an :class:`EmojiSequence` instance from :class:`EmojiCharacter` object or list

        Args:
            value: Single or iterable object of :class:`EmojiCharacter`, composing the sequence

        Returns:
            Instance from internal dictionary

        Raises:
            KeyError: When passed-in value not found in internal dictionary
            TypeError: When passed-in value is not :class:`.EmojiCharacter` object or list
        """
        if isinstance(value, EmojiCharacter):
            s = value.string
        elif isinstance(value, Iterable):
            s = "".join(m.string for m in value)
        else:
            raise TypeError("Argument `value` must be one of `EmojiCharacter` or `Iterable[EmojiCharacter]`")
        return cls.from_string(s)

    @classmethod
    def from_hex(cls, value: Union[int, str, Iterable[Union[int, str]]]) -> Self:
        """Get an :class:`EmojiSequence` instance by unicode code point(s)

        Args:
            value: A single or sequence of HEX string/code.

                it could be:

                - one or more code-point(s) in HEX format string, separated by spaces
                - a single code-point integer
                - An iterable object whose members are code-point string in HEX format
                - An iterable object whose members are code-point integer

        Returns:
            Instance returned from the class's internal dictionary

        Raises:
            KeyError: When passed-in value not found in the class internal dictionary
        """
        cps_array: Iterable[Union[int, str]]
        if isinstance(value, str):
            cps_array = value.split()
        elif isinstance(value, int):
            cps_array = (value,)
        elif isinstance(value, Iterable):
            cps_array = value
        else:
            raise TypeError(
                f"Argument `value` expects to be one of `str`, `bytes`, `int`, or a sequence of that, but actual is {type(value)}"
            )
        return cls.from_characters(EmojiCharacter.from_hex(cp) for cp in cps_array)  # pyright: ignore[reportReturnType]

    @property
    def type_field(self) -> str:
        """A convenience for parsing the emoji sequence files, and is not intended to be maintained as a property.

        may be one of:

        - `"Basic_Emoji"`
        - `"Emoji_Keycap_Sequence"`
        - `"Emoji_Flag_Sequence"`
        - `"Emoji_Tag_Sequence"`
        - `"Emoji_Modifier_Sequence"`
        - `"RGI_Emoji_ZWJ_Sequence"`
        """
        return self._type_field

    @property
    def description(self) -> str:
        """Description"""
        return self._description

    @property
    def version(self) -> str:
        """Version of the Emoji.

        Example:
            ``E0.0``, ``E0.6``, ``E11.0``
        """
        return self._version

    @property
    def variation(self) -> str:
        """``"emoji style"`` or ``"text style"`` of a variable sequence"""
        return self._variation

    @property
    def characters(self) -> Sequence[EmojiCharacter]:
        """Emoji character objects list which makes up the Emoji Sequence"""
        return self._characters

    @property
    def hex(self) -> str:
        """Python style hex string of each emoji-characters's code-point, separated by spaces

        Example:
            ``"0xa9 0xfe0f"``
        """
        return " ".join(c.hex for c in self.characters)

    @property
    def string(self) -> str:
        """string of the Emoji Sequence"""
        return self._string

    @property
    def regex(self) -> str:
        """Regular expression string of the Emoji Sequence"""
        return self._regex

    @property
    def regex_pattern(self):
        """Compiled regular expression pattern of the Emoji Sequence"""
        return self._regex_pat

    @property
    def code_points(self) -> Sequence[int]:
        """List of unicode integer value of the characters who make up this Emoji Sequence"""
        return self._code_points

    @property
    def code_points_string(self) -> str:
        """Unicode style hex string of each emoji-characters's code-point, separated by spaces

        eg: ``"00A9 FE0F"``
        """
        return " ".join(c.code_point_string for c in self.characters)

    @classmethod
    def find_all(cls, s: str) -> Sequence[Tuple[Self, int, int]]:
        """Find out all emoji sequences in a string, and return them in a list

        Items of the returned list is the same as ``yield`` result of :meth:`find`

        The function equals::

            list(EmojiSequence.find(s))

        or ::

            [x for x in EmojiSequence.find(s)]
        """
        return list(cls.find(s))  # pyright: ignore[reportReturnType]

    @classmethod
    def find(cls, s: str) -> Generator[Tuple[Self, int, int], None, None]:
        """Return an iterator which yields all emoji sequences in a string, without actually storing them all simultaneously.

        Args:
            s: The string to find emoji sequences in it

        Yields:
            : Yields at every matched emoji sequence as a 3-members tuple, whose members are:

                0. The found :class:`.EmojiSequence` object
                1. Begin position of the emoji sequence in the string
                2. End position of the emoji sequence in the string
        """
        for m in cls.pattern.finditer(s):
            yield cls.from_string(m.group()), m.start(), m.end()


EmojiSequence.initial()
